/*

	\file		_plugin_manager.h
	\brief 	Private functions of the PluginManager
	\author	Aram Verstegen

	This header contains the declarations of the private Plugin Manager functions:
  -# PluginInit(argc, argv): Initialize plugin and add it to the list.
  -# get_procsymbol(handle, symbol): Retrieve dlsym symbol from specified dl-handle.
  -# print_eef_plugin(log_level, plugin): Print plugin fields to log with specified log_level.
  -# clean_plugin_list(plugin_list): Clean the linkedlist and close all the dl handles.
  -# get_plugin(plugin_name): Retrieve a plugin from the list by name.

*/
#ifndef _PLUGINMANAGER_H
#define _PLUGINMANAGER_H

#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#if HAVE_FNMATCH_H
#include <fnmatch.h>
#endif
#if HAVE_SYSLOG_H
#include <syslog.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_DLFCN_H
#include <dlfcn.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "eef_library.h"

int                               errsv;
const char*                       local_module_dir;
typedef EES_PL_RC                 (*init_fnc_ptr_t)(int, char**); /*!< init function pointer */
typedef EES_PL_RC                 (*run_fnc_ptr_t)(void);					/*!< run function pointer */
typedef EES_PL_RC                 (*term_fnc_ptr_t)(void);				/*!< run function pointer */

char *                            _global_running_plugin_name;
static eef_plugindl_t *           _global_running_plugin;
int	                          _is_threading;

eef_proc_t                        get_procsymbol(void*, char*);
eef_plugindl_t *                  create_plugin_struct(int argc, char** argv, char* name);

EES_RC                            initialize_plugin(eef_plugindl_t*);
EES_PL_RC                         run_plugin_by_name(char*);
EES_PL_RC                         run_plugin(eef_plugindl_t*);
EES_PL_RC                         term_plugin(char*);

EES_RC                            print_eef_plugin(int, eef_plugindl_t*);
EES_RC                            clean_plugin_list(eef_plugindl_t*);

void                              plugin_manager_make_key(void);


#endif

