#include "eef_plugin.h"
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>



#define SPECIAL_EES_OBLIGATION_LIST_ATTR_ID   "http://authz-interop.org/xacml/attribute/trans-obligation"
#define SPECIAL_EES_OBLIGATION_LIST_DELIMITER "?attrId="
#define MAX_OBLIGATIONS                       42

char * getAttributeIDfromAttributeID(char *);
char * url_decode(char *);



EES_PL_RC set_var_as_string(const char* name, char* value);
EES_PL_RC set_var_as_int(const char* name, int value);
EES_PL_RC set_var_as_void_p(const char* name, void* value, size_t size);
char* get_var_as_string(char* name);
int get_var_as_int(char* name);
void* get_var_as_void_p(char* name);


EES_PL_RC plugin_initialize(int argc, char* argv[]){
    aos_dump_argslist();

    return EES_PL_SUCCESS;
}


EES_PL_RC plugin_terminate(){
    return EES_PL_SUCCESS;
}


EES_PL_RC plugin_run(){
    aos_context_t    *_context            = NULL;
    aos_attribute_t  *_tmp_attr           = NULL;
    aos_context_t    *_context_oblig      = NULL;
    aos_attribute_t  *_tmp_attr_oblig     = NULL;

    char             *_tmp_attr_id        = NULL;
    int               i                   = 0;

    /* TODO: needs to change */
    char ** list_of_obligations =  NULL;
    int     list_of_obligations_cnt = 0;
    list_of_obligations = malloc (sizeof (char *) * MAX_OBLIGATIONS);


    EEF_log(LOG_INFO, "Running %s\n", EEF_getRunningPluginName());

    /* Debug */
    aos_dump_argslist();

    /* Search for a special Environment attributeId
     * It's used as a list of Obligation Ids to search for and transform */
    rewindContexts();
    while((_context = getNextContext(ENVIRONMENT, NULL))){
        rewindAttributes(_context);
        while((_tmp_attr = getNextAttribute(_context))){
            _tmp_attr_id = getAttributeId(_tmp_attr);
            if(_tmp_attr_id){
                /* String match beginning of the attributeID with the SPECIAL_EES_OBLIGATION_LIST_ATTR_ID to list its values */
                if(strncmp(_tmp_attr_id, SPECIAL_EES_OBLIGATION_LIST_ATTR_ID, strlen(SPECIAL_EES_OBLIGATION_LIST_ATTR_ID)) == 0){
                    /* List Obligation ID to fetch */
                    list_of_obligations[list_of_obligations_cnt] = getAttributeValueAsString (_tmp_attr);
                    list_of_obligations_cnt++;

                    /* TODO: remove limitation */
                    if(list_of_obligations_cnt == MAX_OBLIGATIONS){
                        EEF_log(LOG_INFO, "Too many obligations to transform, sorry, going to fail. (hardcoded max is %d)\n", MAX_OBLIGATIONS);
                        goto error_transformer_bailout;
                    }
                }
            }
        }
    }


    /* Search the Environment attributes for encoded obligations to transform into real oblgiations
     * attributes to transform to Obligations, when found, register the obligations. */
    rewindContexts();
    while((_context = getNextContext(ENVIRONMENT, NULL))){
        rewindAttributes(_context);
        while((_tmp_attr = getNextAttribute(_context))){
            _tmp_attr_id = getAttributeId(_tmp_attr);
            if(_tmp_attr_id){
                /* Loop through explicitly selected Obligation IDs */
                for (i = 0; i < list_of_obligations_cnt; i++){
                    /*
                     * Compare currently selected attributeID with a listed Obligation ID
                     * The left-hand side of the attribute ID must match, as it ALSO contains the Obligation's attribute ID in URLencoded form
                     * Example:  "http://namespace.tld/xacml/obligation?attribid=http%3A%2F%2Fnamespace.tld%2Fxacml%2Fattribute%2Fexample"
                     * Oblig id: "http://namespace.tld/xacml/obligation"
                     * Attr  id: "http://namespace.tld/xacml/attribute/example"
                     */
                    if(strncmp(_tmp_attr_id, list_of_obligations[i], strlen(list_of_obligations[i])) == 0){
                        /* TODO: Each attribute will have its own obligation context, this must be optimized */
                        if((_context_oblig = createContext(OBLIGATION)) != NULL){
                            char * new_oblig_id = NULL;

                            new_oblig_id        = malloc(sizeof (char) * (strlen (list_of_obligations[i]) + 1)); /* Don't free here, will be owned by the AOS */
                            if (!new_oblig_id){
                                EEF_log(LOG_INFO, "Out of memory. Could not allocate %d bytes at the moment.\n", sizeof (char) * (strlen (list_of_obligations[i]) + 1));
                                goto error_transformer_bailout;
                            }

                            setContextObligationId(_context_oblig, new_oblig_id);
                            if((_tmp_attr_oblig = createAttribute())){
                                char * new_attrib_id     = NULL;
                                char * new_attrib_issuer = NULL;
                                char * new_attrib_type   = NULL;

                                /* Let getAttributeIDfromAttributeID() strip the _tmp_attr_id and 
                                 * return the (URL encoded) right hand side, which contains the 
                                 * attribute ID */
                                new_attrib_id = getAttributeIDfromAttributeID(_tmp_attr_id);
                                if (!new_attrib_id){
                                    EEF_log(LOG_INFO, "Could not strip the Obligation's attribute ID from the Environment attribute ID.\n");
                                    goto error_transformer_bailout;
                                }
                                setAttributeId(_tmp_attr_oblig, new_attrib_id);

                                /* Set issuer in the new attribute */
                                new_attrib_issuer = strdup(getAttributeIssuer(_tmp_attr));
                                if (!new_attrib_issuer){
                                    EEF_log(LOG_INFO, "Out of memory. Could not allocate %d bytes at the moment.\n", sizeof (char) * (strlen (getAttributeIssuer(_tmp_attr)) + 1));
                                    goto error_transformer_bailout;
                                }
                                setAttributeIssuer(_tmp_attr_oblig, new_attrib_issuer);

                                /* Set type in the new attribute */
                                new_attrib_type   = strdup(getAttributeType(_tmp_attr));;
                                if (!new_attrib_type){
                                    EEF_log(LOG_INFO, "Out of memory. Could not allocate %d bytes at the moment.\n", sizeof (char) * (strlen (getAttributeType(_tmp_attr)) + 1));
                                    goto error_transformer_bailout;
                                }
                                setAttributeType(_tmp_attr_oblig, new_attrib_type);

                                /* Data by pointer */
                                setAttributeValue(_tmp_attr_oblig, getAttributeValueAsVoidPointer(_tmp_attr), 0);

                                /* Add attribute to Obligation */
                                addAttribute(_context, _tmp_attr_oblig);
                            }
                            addContext(_context);

                            /* TODO: Remove Environment attribute that got converted. */
                        }
                    }
                }
            }
        }
    }


    free(list_of_obligations);
    return EES_PL_SUCCESS;

error_transformer_bailout:
    free(list_of_obligations);
    return EES_PL_FAILURE;
}


/* getAttributeIDfromAttributeID() will strip the attribute ID at the SPECIAL_EES_OBLIGATION_LIST_DELIMITER
 * followed by urldecoding the string on the right hand side of the delimiter and return it as allocated space
 */
char * getAttributeIDfromAttributeID(char * compoundID){
    char * attr_id = NULL;

    if (compoundID){
        EEF_log(LOG_INFO, "The function %s has nothing to do, empty input.", __func__);
        return NULL;
    }

    /* Strip */
    if (!(attr_id = strstr(compoundID, SPECIAL_EES_OBLIGATION_LIST_DELIMITER))){
        EEF_log(LOG_INFO, "No need to convert anything. No occurance found of %s\n", SPECIAL_EES_OBLIGATION_LIST_DELIMITER);
        return NULL;
    } else {
        /* Move the pointer beyond SPECIAL_EES_OBLIGATION_LIST_DELIMITER */
        attr_id = &attr_id[strlen(SPECIAL_EES_OBLIGATION_LIST_DELIMITER)];


        /* URL decode */
        if (!attr_id){
            return NULL;
        } else {
            return url_decode(attr_id);
        }
    }
}


char * url_decode(char * encoded){
    int i = 0, e = 0, d = 0;
    int encoded_counter = 0;
    char * decoded = NULL;
    char   encoded_char[3];

    if (!encoded)
        return NULL;

    /* Count ecoded character occurances */
    for (i = 0; i < strlen(encoded); i++){
        if (encoded[i] == '%' && isxdigit(encoded[i + 1]) && isxdigit(encoded[i + 2])){
            encoded_counter++;
        }
    }

    /* Allocate space for the result to fit the decoded form */
    decoded = calloc(strlen(encoded) - 2 * encoded_counter + 1, sizeof(char));

    /* Decode string */
    for (d = 0, e = 0; e < strlen(encoded); d++, e++){
        /* Search for encoded characters */
        if (encoded[e] == '%' && isxdigit(encoded[e + 1]) && isxdigit(encoded[e + 2])){
            encoded_char[0] = encoded[e + 1];
            encoded_char[1] = encoded[e + 2];
            encoded_char[2] = '\0';
            e += 2;

            decoded[d] = strtol(encoded_char, NULL, 16);
        }else{
            decoded[d] = encoded[e];
        }
    }
    return decoded;
}


EES_PL_RC set_var_as_string(const char* name, char* value){
    aos_context_t*   context;
    aos_attribute_t* attribute;
    char             *attribute_name=NULL,*running_plug_name;
    int		     attribute_name_len;

    if(value == NULL){
        /*EEF_log(LOG_ERR, "Value cannot be NULL\n");*/
        return EES_PL_FAILURE;
    }

    rewindContexts();
    if((context = getNextContext(NONE, NULL)) == NULL){
        if ((context = createContext (NONE)) != NULL){
            addContext (context);
        } else {
            EEF_log(LOG_ERR, "Unable to get context\n");
            return EES_PL_FAILURE;
        }
    }

    /* store attribute names in AOS for lookup during run */
    if((attribute = createAttribute())){
	running_plug_name=EEF_getRunningPluginName();
	attribute_name_len=strlen(name)+strlen(running_plug_name)+2; /* +1 for _ +1 for \0 */
	/* attribute_name_len >=2: no size 0 problem for malloc */
	if ((attribute_name=malloc(attribute_name_len))==NULL)
	    return EES_PL_FAILURE;
	
        snprintf(attribute_name, attribute_name_len, "%s_%s", name, running_plug_name);
        setAttributeId(   attribute, attribute_name);
        free(attribute_name);

        setAttributeValue(attribute, (void*) value, strlen(value)+1);
        addAttribute(context, attribute);
    }

    return EES_PL_SUCCESS;
}


EES_PL_RC set_var_as_void_p(const char* name, void* value, size_t size){
    aos_context_t*   context;
    aos_attribute_t* attribute;
    char             *attribute_name=NULL,*running_plug_name;
    int		     attribute_name_len;

    /* store attribute names in AOS for lookup during run */

    rewindContexts();
    while((context = getNextContext(NONE, NULL))){
        if(context){
            break;
        }
    }   
    if(context == NULL){
        context = createContext (NONE);
        addContext (context);
    }

    if(context == NULL){
        EEF_log(LOG_ERR, "Unable to get context\n");
        return EES_PL_FAILURE;
    }

    if((attribute = createAttribute())){
	running_plug_name=EEF_getRunningPluginName();
	attribute_name_len=strlen(name)+strlen(running_plug_name)+2; /* +1 for _ +1 for \0 */
	/* attribute_name_len >=2: no size 0 problem for malloc */
	if ((attribute_name=malloc(attribute_name_len))==NULL)
	    return EES_PL_FAILURE;
	
        snprintf(attribute_name, attribute_name_len, "%s_%s", name, running_plug_name);
        setAttributeId(   attribute, attribute_name);
        free(attribute_name);

        setAttributeValue(attribute, value, size);
        addAttribute(context, attribute);
    }
    return EES_PL_SUCCESS;
}

EES_PL_RC set_var_as_int(const char* name, int value){
    aos_context_t*   context;
    aos_attribute_t* attribute;
    char             *attribute_name=NULL,*running_plug_name;
    int		     attribute_name_len;

    /* store attribute names in AOS for lookup during run */

    rewindContexts();
    while((context = getNextContext(NONE, NULL))){
        if(context){
            break;
        }
    }   
    if(context == NULL){
        context = createContext (NONE);
        addContext (context);
    }

    if(context == NULL){
        EEF_log(LOG_ERR, "Unable to get context\n");
        return EES_PL_FAILURE;
    }

    if((attribute = createAttribute())){
	running_plug_name=EEF_getRunningPluginName();
	attribute_name_len=strlen(name)+strlen(running_plug_name)+2; /* +1 for _ +1 for \0 */
	/* attribute_name_len >=2: no size 0 problem for malloc */
	if ((attribute_name=malloc(attribute_name_len))==NULL)
	    return EES_PL_FAILURE;
	
        snprintf(attribute_name, attribute_name_len, "%s_%s", name, running_plug_name);
        setAttributeId(   attribute, attribute_name);
        free(attribute_name);

        setAttributeValue(attribute, &value, sizeof(int));
        addAttribute(context, attribute);
    }
    return EES_PL_SUCCESS;
}

char* get_var_as_string(char* name){
    char *localname=NULL,*running_plug_name;
    int localname_len;
    aos_context_t  * context;
    aos_attribute_t* attribute;

    running_plug_name=EEF_getRunningPluginName();
    localname_len=strlen(name)+strlen(running_plug_name)+2; /* +1 for _ +1 for \0 */
    /* localname_len >=2: no size 0 problem for malloc */
    if ((localname=malloc(localname_len))==NULL)
	return NULL;
    snprintf(localname, localname_len, "%s_%s", name, running_plug_name);

/*    asprintf(&localname, "%s_%s", name, EEF_getRunningPluginName());*/

    rewindContexts();
    while((context = getNextContext(NONE, NULL))){
        rewindAttributes(context);
        while((attribute = getNextAttribute(context))){
            if(strcmp(getAttributeId(attribute), localname) == 0){
                free(localname);
                return getAttributeValueAsString(attribute);
            }
        }
    }
    free(localname);
    return NULL;
}

int get_var_as_int(char* name){
    char *localname=NULL,*running_plug_name;
    int localname_len;
    aos_context_t  * context;
    aos_attribute_t* attribute;

    running_plug_name=EEF_getRunningPluginName();
    localname_len=strlen(name)+strlen(running_plug_name)+2; /* +1 for _ +1 for \0 */
    /* localname_len >=2: no size 0 problem for malloc */
    if ((localname=malloc(localname_len))==NULL)
	return -1;
    snprintf(localname, localname_len, "%s_%s", name, running_plug_name);

    /* Free this */
/*    asprintf(&localname, "%s_%s", name, EEF_getRunningPluginName());*/

    rewindContexts();
    while((context = getNextContext(NONE, NULL))){
        rewindAttributes(context);
        while((attribute = getNextAttribute(context))){
            if(strcmp(getAttributeId(attribute), localname) == 0){
                free(localname);
                return getAttributeValueAsInt(attribute);
            }
        }
    }
    free(localname);
    return -1;
}

void* get_var_as_void_p(char* name){
    char *localname=NULL,*running_plug_name;
    int localname_len;
    aos_context_t  * context;
    aos_attribute_t* attribute;

    running_plug_name=EEF_getRunningPluginName();
    localname_len=strlen(name)+strlen(running_plug_name)+2; /* +1 for _ +1 for \0 */
    /* localname_len >=2: no size 0 problem for malloc */
    if ((localname=malloc(localname_len))==NULL)
	return NULL;
    snprintf(localname, localname_len, "%s_%s", name, running_plug_name);

    /* Free this */
/*    asprintf(&localname, "%s_%s", name, EEF_getRunningPluginName());*/

    rewindContexts();
    while((context = getNextContext(NONE, NULL))){
        rewindAttributes(context);
        while((attribute = getNextAttribute(context))){
            if(strcmp(getAttributeId(attribute), localname) == 0){
                free(localname);
                return *((void**) getAttributeValueAsVoidPointer(attribute));
            }
        }
    }
    free(localname);
    return NULL;
}

