#include "ees_eics.h"
#include "eef_return_codes.h"
#include "eef_log.h"
#include "eef_polytypes.h"
#include "eef_private_types.h"
#include "eef_library.h"
#include "eef_aos.h"

/************************************************
Function:    extractRequestSubjectToAOS
Parameters:  const xacml_request_t           request
Description:
             This function extracts the Subject information details
             from the request and puts that into the EEF AOS.
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestSubjectToAOS (const xacml_request_t request)
{
  /*char * logstr = "extractRequestSubjectToAOS()";*/
    size_t count = 0;
    size_t i     = 0;

    const char *category;
    const char *attribute_id;
    const char *data_type;
    const char *issuer;
    const char *value;
        
    aos_context_t*   aos_context = NULL;
    aos_attribute_t* aos_attribute = NULL;

    xacml_request_get_subject_attribute_count(request, &count);

    if (count > 0)
    {
      rewindContexts();
      if((aos_context = getNextContext(SUBJECT, NULL)) == NULL){
        if((aos_context = createContext (SUBJECT)) == NULL){
          return 1;
        }
      }
      /* aos_context should be set now */
    }

    for (i = 0; i < count; i++)
    {

        category     = NULL;
        attribute_id = NULL;
        data_type    = NULL;
        issuer       = NULL;
        value        = NULL;

        xacml_request_get_subject_attribute(
                request, i, &category, &attribute_id, &data_type, &issuer, &value);

        /* EEF AOS magic */
        /*fprintf (stderr, "aos_set_string (\"%s\", \"%s\")\n", attribute_id, (const char *) value);*/
        /* if (strcmp ("http://authz-interop.org/xacml/subject/voms-fqan", attribute_id) == 0) */
        /* { */
            /* fprintf (stderr, "Skipping voms-fqan\n"); */
            /* continue; */
        /* } */

        /* aos_set_string (attribute_id, (char *) value); */

        if (aos_context){
          if((aos_attribute = createAttribute())){
            setAttributeId    (aos_attribute, attribute_id);
            setAttributeIssuer(aos_attribute, issuer);
            setAttributeType  (aos_attribute, data_type);
            setAttributeValue (aos_attribute, value, strlen(value) + 1);
            addAttribute(aos_context, aos_attribute);
          }
        }
    }

    if(aos_context){
      addContext (aos_context);
    }

    return 0;
}

/************************************************
Function:    extractRequestActionToAOS
Parameters:  const xacml_request_t           request
Description:
             This function extracts the Action information details
             from the request and puts that into the EEF AOS.
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestActionToAOS (const xacml_request_t request)
{
    int                           rc;
    size_t                        i;
    size_t                        count;
    const char *                  attribute_id;
    const char *                  data_type;
    const char *                  issuer;
    const char *                  value;
    /*char *                        logstr = "extractRequestActionToAOS";*/

    aos_context_t*   aos_context = NULL;
    aos_attribute_t* aos_attribute = NULL;

    rc = xacml_request_get_action_attribute_count(request, &count);


    if (count > 0)
    {
      rewindContexts();
      if((aos_context = getNextContext(ACTION, NULL)) == NULL){
        if((aos_context = createContext (ACTION)) == NULL){
          return 1;
        }
      }
      /* aos_context should be set now */
    }

    for (i = 0 ; i < count; i++)
    {
        attribute_id = NULL;
        data_type    = NULL;
        issuer       = NULL;
        value        = NULL;
        rc = xacml_request_get_action_attribute(
                request,
                i,
                &attribute_id,
                &data_type,
                &issuer,
                &value);
        /* EEF AOS magic */
        /*fprintf (stderr, "aos_set_string (\"%s\", \"%s\")\n", attribute_id, (char *) value);*/

        /* aos_set_string (attribute_id, (char *) value); */

        if (aos_context){
          if((aos_attribute = createAttribute())){
            setAttributeId    (aos_attribute, attribute_id);
            setAttributeIssuer(aos_attribute, issuer);
            setAttributeType  (aos_attribute, data_type);
            setAttributeValue (aos_attribute, value, strlen(value) + 1);
            addAttribute(aos_context, aos_attribute);
          }
        }
    }

    if(aos_context){
      addContext (aos_context);
    }

    return 0;
}

/************************************************
Function:    extractRequestResourceToAOS
Parameters:  const xacml_request_t           request
Description:
             This function extracts the Action information details
             from the request and puts that into the EEF AOS
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestResourceToAOS (const xacml_request_t           request)
{
    int                                 rc;
    size_t                              i;
    size_t                              j;
    size_t                              count;
    const char *                        attribute_id;
    const char *                        data_type;
    const char *                        issuer;
    const char *                        value;
    xacml_resource_attribute_t          resource_attribute  = NULL;
    size_t                              resource_attr_count = 0;
    /*const char *                        logstr = "extractRequestResourceToAOS";*/

    aos_context_t*   aos_context = NULL;
    aos_attribute_t* aos_attribute = NULL;


    rc = xacml_request_get_resource_attribute_count(request, &resource_attr_count);

    for (i = 0 ; i < resource_attr_count; i++)
    {
        rc = xacml_request_get_resource_attribute(
                request,
                i,
                &resource_attribute);

        if (resource_attribute)
        {
            rc = xacml_resource_attribute_get_count(
                       resource_attribute,
                       &count);

            if((aos_context = createContext (RESOURCE)) == NULL){
              return 1;
            }
            /* aos_context should be set now */

            for (j = 0; j < count; j++)
            {
                attribute_id = NULL;
                data_type    = NULL;
                issuer       = NULL;
                value        = NULL;
                rc = xacml_resource_attribute_get_attribute(
                           resource_attribute,
                           j,
                           &attribute_id,
                           &data_type,
                           &issuer,
                           &value);
                /* EEF AOS magic */
                /*fprintf (stderr, "aos_set_string (\"%s\", \"%s\")\n", attribute_id, (char *) value);*/

                /* aos_set_string (attribute_id, (char *) value); */

                if (aos_context){
                  if((aos_attribute = createAttribute())){
                    setAttributeId    (aos_attribute, attribute_id);
                    setAttributeIssuer(aos_attribute, issuer);
                    setAttributeType  (aos_attribute, data_type);
                    setAttributeValue (aos_attribute, value, strlen(value) + 1);
                    addAttribute(aos_context, aos_attribute);
                    aos_attribute = NULL;
                  }
                }
            }
            if(aos_context){
              addContext (aos_context);
              aos_context = NULL;
            }
        }
    }

    return 0;
}

/************************************************
Function:    extractRequestEnvironmentToAOS
Parameters:  const xacml_request_t           request
Description:
             This function extracts the Action information details
             from the request and puts that into the EEF AOS.
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestEnvironmentToAOS (const xacml_request_t request)
{
    int                           rc;
    size_t                        i;
    size_t                        count;
    const char *                        attribute_id;
    const char *                        data_type;
    const char *                        issuer;
    const char *                        value;

    /*char *                        logstr = "extractRequestEnvironmentToAOS";*/

    aos_context_t*   aos_context   = NULL;
    aos_attribute_t* aos_attribute = NULL;

    rc = xacml_request_get_environment_attribute_count(request, &count);


    if (count > 0)
    {
      rewindContexts();
      if((aos_context = getNextContext(ENVIRONMENT, NULL)) == NULL){
        if((aos_context = createContext (ENVIRONMENT)) == NULL){
          return 1;
        }
      }
      /* aos_context should be set now */
    }

    for (i = 0 ; i < count; i++)
    {
        attribute_id = NULL;
        data_type    = NULL;
        issuer       = NULL;
        value        = NULL;
        rc = xacml_request_get_environment_attribute(
                request,
                i,
                &attribute_id,
                &data_type,
                &issuer,
                &value);

        /* EEF AOS magic */
        /*fprintf (stderr, "aos_set_string (\"%s\", \"%s\")\n", attribute_id, (char *) value);*/

        /* aos_set_string (attribute_id, (char *) value); */

        if (aos_context){
          if((aos_attribute = createAttribute())){
            setAttributeId    (aos_attribute, attribute_id);
            setAttributeIssuer(aos_attribute, issuer);
            setAttributeType  (aos_attribute, data_type);
            setAttributeValue (aos_attribute, value, strlen(value) + 1);
            addAttribute(aos_context, aos_attribute);
          }
        }
    }

    if(aos_context){
      addContext (aos_context);
    }

    return 0;
}


/************************************************
Function:    destroyRequestResource
Parameters:  xacml_profile_request_t * xacml_profile_request
Description:
             This function will destroy and free the Resource information details
             from the request, which was set by the extractRequestResource() function
             to a xacml_profile_request_t
Return:
          0 : good
         !0 : bad 
************************************************/
int destroyRequestResource (xacml_profile_request_t * xacml_profile_request) {
  xacml_profile_resource_t profile_resource = xacml_profile_request->xacml_profile_resource;

  if (profile_resource.xacml_resource_x509_id.resource_x509_id){
    free(profile_resource.xacml_resource_x509_id.resource_x509_id);
  }
  if (profile_resource.xacml_resource_x509_issuer.resource_x509_issuer){
    free(profile_resource.xacml_resource_x509_issuer.resource_x509_issuer);
  }
  if (profile_resource.xacml_resource_dns_name.resource_dns_name){
    free(profile_resource.xacml_resource_dns_name.resource_dns_name);
  }
  return 0;
}


/************************************************
Function:    constructFailureResponse
Parameters:  xacml_response_t * response
Description:
             The XACML response message to indicate a failure is constructed here.
************************************************/
int constructFailureResponse (xacml_response_t * response) {
  xacml_response_set_saml_status_code  (*response, SAML_STATUS_AuthnFailed);
  xacml_response_set_xacml_status_code (*response, XACML_STATUS_ok);
  xacml_response_set_xacml_decision    (*response, XACML_DECISION_Deny);

  return 0;
}



/************************************************
Function:    constructResponse
Parameters:  xacml_response_t * response
Description:
             The XACML response message is constructed here.
             This implementation will use the Unix UID, Primary GID and
             multiple Secondary GIDs as input to contruct the obligations and
             its attributes. 
************************************************/
int constructResponse (xacml_response_t * response){

  /* Mapping Information Translated */
  aos_context_t*       context         = NULL;
  aos_attribute_t*     attribute       = NULL;
  xacml_obligation_t   tmp_obligation  = NULL;
  char*                attribute_name  = NULL;
  char*                attribute_value = NULL;
  char*                obligation_id   = NULL;

  rewindContexts();
  while((context = getNextContext(OBLIGATION, NULL)) != NULL){
    rewindAttributes(context);
    obligation_id = getContextObligationId(context),
    xacml_obligation_init(&tmp_obligation,
                          obligation_id,
                          XACML_EFFECT_Permit);
    /*printf("Obligation: %s at %p", obligation_id, tmp_obligation);*/
    /*EEF_log(LOG_DEBUG, "Obligation %s", obligation_id);*/
    while((attribute = getNextAttribute(context)) != NULL){
      attribute_name = getAttributeId(attribute);
      attribute_value = getAttributeValueAsString(attribute);
      if(attribute_name && attribute_value){
        /*EEF_log(LOG_DEBUG, "\t%s=%s\n", attribute_name, attribute_value);*/

        /* uid */
        xacml_obligation_add_attribute(tmp_obligation,
                                       attribute_name,
                                       XACML_DATATYPE_STRING,
                                       attribute_value);
        /*printf("Added obligation at: %p type %s\n", tmp_obligation, XACML_DATATYPE_STRING);*/
      }   
    }   
    xacml_response_add_obligation(*response, tmp_obligation);
    xacml_obligation_destroy(tmp_obligation);
    tmp_obligation = NULL;
  }

  /*********** E: Obligation UIDGID ***********/


  xacml_response_set_saml_status_code  (*response, SAML_STATUS_Success);
  xacml_response_set_xacml_status_code (*response, XACML_STATUS_ok);
  xacml_response_set_xacml_decision    (*response, XACML_DECISION_Permit);

  return 0;
}

int eef_xacml_authorize(void *handler_arg, const xacml_request_t request, xacml_response_t response) {
  pid_t             tid;
  /*char * logstr = "xacml_authorize()";*/
  /*int 	 rc     = 0;*/
  /*xacml_status_code_t my_decision = XACML_STATUS_ok;*/

  xacml_response_set_issuer(response, "EES");

  extractRequestSubjectToAOS (request);
  extractRequestActionToAOS (request);
  extractRequestResourceToAOS (request);
  extractRequestEnvironmentToAOS (request);

  tid = syscall(SYS_gettid);
  EEF_log(LOG_DEBUG, "In thread %i\n", tid);

  if(EEF_run() == EES_FAILURE) { 
    constructFailureResponse(&response);
    EEF_log(LOG_ERR, "Failed to run the EEF successfully\n");
  } else {
    constructResponse(&response);
  }

  return 0;
}
